config = {
    # We need to tell the fedmsg-hub that it should load our consumer on start.
    "fedmsg.consumers.badges.enabled": True,

    # This tells the consumer where to look for its BadgeRule definitions.  It
    # may be a relative or an absolute path on the file system.
    "badges.yaml.directory": "tests/test_badges",

    # Number of seconds to delay before consuming a message for our event loop.
    # This is here to help us mitigate distributed race conditions between
    # fedbadges and datanommer.
    "badges.consume_delay": 1,

    # This is a dictionary of tahrir-related configuration
    "badges_global": {

        # This is a sqlalchemy URI that points to the tahrir DB.  In
        # production, this will be a postgres URI.
        "database_uri": "sqlite:////tmp/test-badges-sqlite.db",

        # This is a set of data that tells our consumer what Open Badges Issuer
        # should be kept as the issuer of all the badges we create.
        "badge_issuer": dict(
            issuer_id='Fedora Project',
            issuer_origin='http://badges.fedoraproject.com',
            issuer_name='Fedora Project',
            issuer_org='http://fedoraproject.org',
            issuer_contact='badges@fedoraproject.org',
        ),
    },

    "fedbadges.datagrepper_url": "https://apps.fedoraproject.org/datagrepper",
}



# -*- coding; utf-8 -*-
# Author: Ross Delinger
# Description: A system to award Fedora Badges (open badges)
# based on messages on the bu

from fedmsg.commands import BaseCommand
from fedbadges.consumers import FedoraBadgesConsumer


class BadgesCommand(BaseCommand):
    """ Relay connections to the bus, and enabled the badges consumer """
    name = 'fedmsg-badges'
    extra_args = []
    daemonizable = True

    def run(self):
        moksha_options = dict(
            zmq_subscribe_endpoints=','.join(
                ','.join(bunch) for bunch in self.config['endpoints'].values()
            ),
        )
        self.config.update(moksha_options)
        self.config['fedmsg.consumers.badges.enabled'] = True

        from moksha.hub import main
        main(
            options=self.config,
            # If you omit this argument, it will pick up *all* consumers.
            consumers=[FedoraBadgesConsumer],
            producers=[],
        )


def badges():
    command = BadgesCommand()
    command.execute()



# -*- coding; utf-8 -*-
""" Models for fedbadges.

The primary thing here is a "BadgeRule" which is an in-memory working
abstraction of the trigger and criteria required to award a badge.

Authors:    Ralph Bean
"""

import abc
import copy
import json
import types
import functools
import inspect
import transaction
import re

import fedmsg.config
import fedmsg.meta
import fedmsg.encoding
import datanommer.models

from fedbadges.utils import (
    # These are all in-process utilities
    construct_substitutions,
    format_args,
    single_argument_lambda_factory,
    recursive_lambda_factory,
    graceful,

    # These make networked API calls
    user_exists_in_fas,
)

import logging
log = logging.getLogger('moksha.hub')


nick2fas = None
try:
    from fedmsg_meta_fedora_infrastructure.fasshim import nick2fas
except ImportError as e:
    log.warn("Could not import nick2fas: %r" % e)


email2fas = None
try:
    from fedmsg_meta_fedora_infrastructure.fasshim import email2fas
except ImportError as e:
    log.warn("Could not import email2fas: %r" % e)

# Match OpenID agent strings, i.e. http://FAS.id.fedoraproject.org
def openid2fas(openid, **config):
    m = re.search('^https?://([a-z][a-z0-9]+)\.id\.fedoraproject\.org$', openid)
    if m:
        return m.group(1)
    return openid

operators = frozenset([
    "all",
    "any",
    "not",
])
lambdas = frozenset([
    "lambda",
])

operator_lookup = {
    "any": any,
    "all": all,
    "not": lambda x: all([not item for item in x])
}

fedmsg_config = fedmsg.config.load_config()
fedmsg.meta.make_processors(**fedmsg_config)


class BadgeRule(object):
    required = frozenset([
        'name',
        'image_url',
        'description',
        'creator',
        'discussion',
        'issuer_id',

        'trigger',
        'criteria',
    ])

    possible = required.union([
        'recipient',
        'recipient_nick2fas',
        'recipient_email2fas',
        'recipient_openid2fas',
    ])

    banned_usernames = frozenset([
        'bodhi',
        'oscar',
        'apache',
        'koji',
        'bodhi',
        'taskotron',
    ])

    def __init__(self, badge_dict, tahrir_database, issuer_id):
        argued_fields = frozenset(badge_dict.keys())

        if not argued_fields.issubset(self.possible):
            raise KeyError(
                "%r are not possible fields.  Choose from %r" % (
                    argued_fields.difference(self.possible),
                    self.possible
                ))

        if not self.required.issubset(argued_fields):
            raise ValueError(
                "BadgeRule requires %r.  Missing %r" % (
                    self.required,
                    self.required.difference(argued_fields),
                ))

        self._d = badge_dict

        self.tahrir = tahrir_database
        if self.tahrir:
            transaction.begin()
            self.badge_id = self._d['badge_id'] = self.tahrir.add_badge(
                name=self._d['name'],
                image=self._d['image_url'],
                desc=self._d['description'],
                criteria=self._d['discussion'],
                tags=','.join(self._d.get('tags', [])),
                issuer_id=issuer_id,
            )
            transaction.commit()

        self.trigger = Trigger(self._d['trigger'], self)
        self.criteria = Criteria(self._d['criteria'], self)
        self.recipient_key = self._d.get('recipient')
        self.recipient_nick2fas = self._d.get('recipient_nick2fas')
        self.recipient_email2fas = self._d.get('recipient_email2fas')
        self.recipient_openid2fas = self._d.get('recipient_openid2fas')

        # A sanity check before we kick things off.
        if self.recipient_nick2fas and not nick2fas:
            raise ImportError("recipient_nick2fas specified, but "
                              "nick2fas is not available.")

        if self.recipient_email2fas and not email2fas:
            raise ImportError("recipient_email2fas specified, but "
                              "email2fas is not available.")

    def __getitem__(self, key):
        return self._d[key]

    def __repr__(self):
        return "<fedbadges.models.BadgeRule: %r>" % self._d

    def matches(self, msg):

        # First, do a lightweight check to see if the msg matches a pattern.
        if not self.trigger.matches(msg):
            return set()

        # Before proceeding further, let's see who would get this badge if
        # our more heavyweight checks matched up.  If the user specifies a
        # recipient_key, we can use that to extract the potential awardee.  If
        # that is not specified, we just use `msg2usernames`.
        if self.recipient_key:
            subs = construct_substitutions(msg)
            obj = format_args(self.recipient_key, subs)

            if isinstance(obj, (basestring, int, float)):
                obj = [obj]

            # It is possible to recieve a list of dictionary containing the name
            # of the recipient, this is the case in the pagure's fedmsg.
            # In that case we create a new list containing the names taken from the
            # dictionnary.
            # If the "name" was not available in the dictionary we raise an exception.
            new_obj = []
            for item in obj:
                if isinstance(item, dict):
                    try:
                        new_obj.append(item["name"])
                    except KeyError:
                        raise Exception("Multiple recipients : name not found in the message")
            if new_obj:
                obj = new_obj


            # On the way, it is possible for the fedmsg message to contain None
            # for "agent".  A problem here though is that None is not iterable,
            # so let's replace it with an equivalently empty iterable so code
            # further down doesn't freak out.  An instance of this is when a
            # user without a fas account comments on a bodhi update.
            if obj is None:
                obj = []

            awardees = frozenset(obj)

            if self.recipient_nick2fas:
                awardees = frozenset([
                    nick2fas(nick, **fedmsg_config) for nick in awardees
                ])

            if self.recipient_email2fas:
                awardees = frozenset([
                    email2fas(email, **fedmsg_config) for email in awardees
                ])

            if self.recipient_openid2fas:
                awardees = frozenset([
                    openid2fas(openid, **fedmsg_config) for openid in awardees
                ])
        else:
            usernames = fedmsg.meta.msg2usernames(msg)
            awardees = usernames.difference(self.banned_usernames)

        # Strip anyone who is an IP address
        awardees = frozenset([
            user for user in awardees if not (
                user.startswith('192.168.') or
                user.startswith('10.')
            )
        ])

        # If no-one would get the badge by default, then no reason to waste
        # time doing any further checks.  No need to query the Tahrir DB.
        if not awardees:
            return awardees

        # Limit awardees to only those who do not already have this badge.
        # Do this only if we have an active connection to the Tahrir DB.
        if self.tahrir:
            awardees = frozenset([
                user for user in awardees
                if not self.tahrir.assertion_exists(
                    self.badge_id, "%s@fedoraproject.org" % user
                )])

            # Also, exclude any potential awardees who have opted out.
            awardees = frozenset([
                user for user in awardees
                if not self.tahrir.person_opted_out(
                    "%s@fedoraproject.org" % user
                )])

        # If no-one would get the badge at this point, then no reason to waste
        # time doing any further checks.  No need to query datanommer.
        if not awardees:
            return awardees

        # Check our backend criteria -- likely, perform datanommer queries.
        try:
            if not self.criteria.matches(msg):
                return set()
        except IOError as e:
            log.exception(e)
            return set()

        # Lastly, and this is probably most expensive.  Make sure the person
        # actually has a FAS account before we award anything.
        # https://github.com/fedora-infra/tahrir/issues/225
        awardees = set([
            u for u in awardees if user_exists_in_fas(fedmsg_config, u)
        ])

        return awardees


class AbstractComparator(object):
    """ Base class for shared behavior between trigger and criteria. """
    __metaclass__ = abc.ABCMeta
    possible = required = frozenset()
    children = None

    def __init__(self, d, parent=None):
        argued_fields = frozenset(d.keys())
        if not argued_fields.issubset(self.possible):
            raise KeyError(
                "%r are not possible fields.  Choose from %r" % (
                    argued_fields.difference(self.possible),
                    self.possible
                ))

        if self.required and not self.required.issubset(argued_fields):
            raise ValueError(
                "%r are required fields.  Missing %r" % (
                    self.required,
                    self.required.difference(argued_fields),
                ))

        self._d = d
        self.parent = parent

    def __repr__(self):
        return "<%s: %r> which is a child of %s" % (
            type(self).__name__, self._d, repr(self.parent)
        )

    @abc.abstractmethod
    def matches(self, msg):
        pass


class AbstractTopLevelComparator(AbstractComparator):
    def __init__(self, *args, **kwargs):
        super(AbstractTopLevelComparator, self).__init__(*args, **kwargs)
        cls = type(self)

        if len(self._d) > 1:
            raise ValueError("No more than one trigger allowed.  "
                             "Use an operator, one of %r" % operators)

        self.attribute = self._d.keys()[0]
        self.expected_value = self._d[self.attribute]

        # XXX - Check if we should we recursively nest Trigger/Criteria?

        # First, trick negation into thinking it is not a unary operator.
        if self.attribute == 'not':
            self.expected_value = [self.expected_value]

        # Then, treat everything as if it accepts an arbitrary # of args.
        if self.attribute in operators:
            if not isinstance(self.expected_value, list):
                raise TypeError("Operators only accept lists, not %r" %
                                type(self.expected_value))
            self.children = [cls(child, self) for child in self.expected_value]


class Trigger(AbstractTopLevelComparator):
    possible = frozenset([
        'topic',
        'category',
    ]).union(operators).union(lambdas)

    @graceful(set())
    def matches(self, msg):
        # Check if we should just aggregate the results of our children.
        # Otherwise, we are a leaf-node doing a straightforward comparison.
        if self.children:
            return operator_lookup[self.attribute]((
                child.matches(msg) for child in self.children
            ))
        elif self.attribute == 'lambda':
            return single_argument_lambda_factory(
                expression=self.expected_value, argument=msg, name='msg')
        elif self.attribute == 'category':
            # TODO -- use fedmsg.meta.msg2processor(msg).__name__.lower()
            return msg['topic'].split('.')[3] == self.expected_value
        else:
            if hasattr(msg[self.attribute], 'endswith'):
                return msg[self.attribute].endswith(self.expected_value)
            else:
                return msg[self.attribute] == self.expected_value


class Criteria(AbstractTopLevelComparator):
    possible = frozenset([
        'datanommer',
    ]).union(operators)

    def __init__(self, *args, **kwargs):
        super(Criteria, self).__init__(*args, **kwargs)

        if not self.children:
            # Then, by AbstractComparator rules, I am a leaf node.  Specialize!
            self._specialize()

    def _specialize(self):
        if self.attribute == 'datanommer':
            self.specialization = DatanommerCriteria(self.expected_value)
        # TODO -- expand this with other "backends" as necessary
        # elif self.attribute == 'fas'
        else:
            raise RuntimeError("This should be impossible to reach.")

    @graceful(set())
    def matches(self, msg):
        if self.children:
            return operator_lookup[self.attribute]((
                child.matches(msg) for child in self.children
            ))
        else:
            return self.specialization.matches(msg)


class AbstractSpecializedComparator(AbstractComparator):
    pass

class DatanommerCriteria(AbstractSpecializedComparator):
    required = possible = frozenset([
        'filter',
        'operation',
        'condition',
    ])

    condition_callbacks = {
        'is greater than or equal to': lambda t, v: v >= t,
        'greater than or equal to': lambda t, v: v >= t,
        'greater than': lambda t, v: v > t,

        'is less than or equal to': lambda t, v: v <= t,
        'less than or equal to': lambda t, v: v <= t,
        'less than': lambda t, v: v < t,

        'equal to': lambda t, v: v == t,
        'is equal to': lambda t, v: v == t,

        'is not': lambda t, v: v != t,
        'is not equal to': lambda t, v: v != t,

        'lambda': single_argument_lambda_factory,
    }

    def __init__(self, *args, **kwargs):
        super(DatanommerCriteria, self).__init__(*args, **kwargs)
        if len(self._d['condition']) > 1:
            conditions = self.condition_callbacks.keys()
            raise ValueError("No more than one condition allowed.  "
                             "Use one of %r" % conditions)

        # Determine what arguments datanommer..grep accepts
        argspec = inspect.getargspec(datanommer.models.Message.grep)
        irrelevant = frozenset(['defer'])
        grep_arguments = frozenset(argspec.args[1:]).difference(irrelevant)

        # Validate the filter
        argued_filter_fields = frozenset(self._d['filter'].keys())
        if not argued_filter_fields.issubset(grep_arguments):
            raise KeyError(
                "%r are not possible fields.  Choose from %r" % (
                    argued_filter_fields.difference(grep_arguments),
                    grep_arguments,
                ))

        # Validate the condition
        condition_key, condition_val = self._d['condition'].items()[0]
        if condition_key not in self.condition_callbacks:
            raise KeyError("%r is not a valid condition key.  Use one of %r" %
                           (condition_key, self.condition_callbacks.keys()))

        # Construct a condition callable for later
        self.condition = functools.partial(
            self.condition_callbacks[condition_key], condition_val)

    def _construct_query(self, msg):
        """ Construct a datanommer query for this message.

        The "filter" section of this criteria object will be used.  It will
        first be formatted with any substitutions present in the incoming
        message.  This is used, for instance, to construct a query like "give
        me all the messages bearing the same topic as the message that just
        arrived".
        """
        subs = construct_substitutions(msg)
        kwargs = format_args(copy.copy(self._d['filter']), subs)
        kwargs = recursive_lambda_factory(kwargs, msg, name='msg')
        kwargs['defer'] = True
        total, pages, query = datanommer.models.Message.grep(**kwargs)
        return total, pages, query

    def _format_lambda_operation(self, msg):
        """ Format the string representation of a lambda operation.

        The lambda operation can be formatted here to include strings that
        appear in the message being evaluated like
        %(msg.comment.update_submitter)s.  Placeholders like that will have
        their value substituted with whatever appears in the incoming message.
        """
        subs = construct_substitutions(msg)
        operation = format_args(copy.copy(self._d['operation']), subs)
        return operation['lambda']

    def matches(self, msg):
        """ A datanommer criteria check is composed of three steps.

        - A datanommer query is constructed by combining our yaml definition
          with the incoming fedmsg message that triggered us.
        - An operation in python is constructed by comining our yaml definition
          with the incoming fedmsg message that triggered us.  That operation
          is then executed against the datanommer query object.
        - A condition, derived from our yaml definition, is evaluated with the
          result of the operation from the previous step and is returned.
        """
        total, pages, query = self._construct_query(msg)
        if self._d['operation'] == 'count':
            result = total
        elif isinstance(self._d['operation'], dict):
            expression = self._format_lambda_operation(msg)
            result = single_argument_lambda_factory(
                expression=expression, argument=query, name='query')
        else:
            operation = getattr(query, self._d['operation'])
            result = operation()
        return self.condition(result)



""" Utilities for fedbadges that don't quite fit anywhere else. """

import types

import logging
log = logging.getLogger("moksha.hub")

import fedmsg
import fedora.client
import requests

# These are here just so they're available in globals()
# for compiling lambda expressions
import json
import re
import fedmsg.config
import fedmsg.encoding
import fedmsg.meta


def construct_substitutions(msg):
    """ Convert a fedmsg message into a dict of substitutions. """
    subs = {}
    for key1 in msg:
        if isinstance(msg[key1], dict):
            subs.update(dict([
                ('.'.join([key1, key2]), val2)
                for key2, val2 in construct_substitutions(msg[key1]).items()
            ]))
            subs[key1] = msg[key1]
        elif isinstance(msg[key1], basestring):
            subs[key1] = msg[key1].lower()
        else:
            subs[key1] = msg[key1]
    return subs


def format_args(obj, subs):
    """ Recursively apply a substitutions dict to a given criteria subtree """

    if isinstance(obj, dict):
        for key in obj:
            obj[key] = format_args(obj[key], subs)
    elif isinstance(obj, list):
        return [format_args(item, subs) for item in obj]
    elif isinstance(obj, basestring) and obj[2:-2] in subs:
        obj = subs[obj[2:-2]]
    elif isinstance(obj, (int, float)):
        pass
    else:
        obj = obj % subs

    return obj


def single_argument_lambda_factory(expression, argument, name='value'):
    """ Compile and execute a lambda expression with a single argument """

    code = compile("lambda %s: %s" % (name, expression), __file__, "eval")
    func = types.LambdaType(code, globals())()
    return func(argument)


def recursive_lambda_factory(obj, arg, name='value'):
    """ Given a dict, find any lambdas, compile, and execute them. """

    if isinstance(obj, dict):
        for key in obj:
            if key == 'lambda':
                # If so, *replace* the parent dict with the result of the expr
                obj = single_argument_lambda_factory(obj[key], arg, name)
                break
            else:
                obj[key] = recursive_lambda_factory(obj[key], arg, name)
    elif isinstance(obj, list):
        return [recursive_lambda_factory(e, arg, name) for e in obj]
    else:
        pass

    return obj


def graceful(default_return_value):
    """ A decorator that gracefully handles exceptions. """

    def decorate(method):
        def inner(self, *args, **kwargs):
            try:
                return method(self, *args, **kwargs)
            except Exception as e:
                log.exception(e)
                log.error("From method: %r self: %r args: %r kwargs: %r" % (
                    method, self, args, kwargs))
                return default_return_value
        return inner
    return decorate


def notification_callback(topic, msg):
    """ This is a callback called by tahrir_api whenever something
    it deems important has happened.

    It is just used to publish fedmsg messages.
    """
    fedmsg.publish(
        topic=topic,
        msg=msg,
    )


def user_exists_in_fas(config, user):
    """ Return true if the user exists in FAS. """
    default_url = 'https://admin.fedoraproject.org/accounts/'
    fas2 = fedora.client.AccountSystem(
        base_url=config['fas_credentials'].get('base_url', default_url),
        username=config['fas_credentials']['username'],
        password=config['fas_credentials']['password'],
    )
    return bool(fas2.person_by_username(user))



from setuptools import setup

# Need this to get around weird setuptools biz on py2.7
try:
    import multiprocessing
    import logging
except ImportError:
    pass


def get_requirements(filename='requirements.txt'):
    with open(filename, 'r') as f:
        return [line for line in f.readlines()]


long_description = open("README.rst").read().strip()


setup(
    name='fedbadges',
    version='1.0.0',
    description='fedmsg consumer for awarding open badges',
    long_description=long_description,
    license='GPLv2+',
    author='Ross Delinger',
    author_email='rdelinge@redhat.com',
    maintainer='Ralph Bean',
    maintainer_email='rbean@redhat.com',
    url='https://github.com/fedora-infra/fedbadges',
    install_requires=get_requirements(),
    tests_require=[
        "nose",
        "mock",
    ],
    test_suite="nose.collector",
    include_package_data=True,
    packages=[
        'fedbadges',
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            "fedmsg-badges=fedbadges.commands:badges"
        ],
        'moksha.consumer': [
            "fedmsg-badges=fedbadges.consumers:FedoraBadgesConsumer",
        ]
    },
)



# -*- coding; utf-8 -*-
""" Consumers for producing openbadges by listening for messages on fedmsg

Authors:  Ross Delinger
          Ralph Bean
"""

import os.path
import yaml
import traceback
import transaction
import threading
import time

import fedmsg.consumers

import tahrir_api.dbapi
import datanommer.models

from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker, scoped_session
from zope.sqlalchemy import ZopeTransactionExtension

import fedbadges.rules
import fedbadges.utils

import logging
log = logging.getLogger("moksha.hub")


class FedoraBadgesConsumer(fedmsg.consumers.FedmsgConsumer):
    topic = "*"
    config_key = "fedmsg.consumers.badges.enabled"
    consume_delay = 3
    delay_limit = 100

    def __init__(self, hub):
        self.badge_rules = []
        self.hub = hub

        super(FedoraBadgesConsumer, self).__init__(hub)

        self.consume_delay = int(self.hub.config.get('badges.consume_delay',
                                                     self.consume_delay))
        self.delay_limit = int(self.hub.config.get('badges.delay_limit',
                                                   self.delay_limit))

        # Five things need doing at start up time
        # 0) Set up a request local to hang thread-safe db sessions on.
        # 1) Initialize our connection to the tahrir DB and perform some
        #    administrivia.
        # 2) Initialize our connection to the datanommer DB.
        # 3) Load our badge definitions and rules from YAML.
        # 4) Initialize fedmsg so that those listening to us can handshake.

        # Thread-local stuff
        self.l = threading.local()

        # Tahrir stuff.
        self._initialize_tahrir_connection()

        # Datanommer stuff
        self._initialize_datanommer_connection()

        # Load badge definitions
        directory = hub.config.get("badges.yaml.directory", "badges_yaml_dir")
        self.badge_rules = self._load_badges_from_yaml(directory)

    def _initialize_tahrir_connection(self):
        if hasattr(self.l, 'tahrir'):
            return

        global_settings = self.hub.config.get("badges_global", {})

        database_uri = global_settings.get('database_uri')
        if not database_uri:
            raise ValueError('Badges consumer requires a database uri')

        session_cls = scoped_session(sessionmaker(
            extension=ZopeTransactionExtension(),
            bind=create_engine(database_uri),
        ))

        self.l.tahrir = tahrir_api.dbapi.TahrirDatabase(
            session=session_cls(),
            autocommit=False,
            notification_callback=fedbadges.utils.notification_callback,
        )
        issuer = global_settings.get('badge_issuer')

        transaction.begin()
        self.issuer_id = self.l.tahrir.add_issuer(
            issuer.get('issuer_origin'),
            issuer.get('issuer_name'),
            issuer.get('issuer_org'),
            issuer.get('issuer_contact')
        )
        transaction.commit()

    def _initialize_datanommer_connection(self):
        datanommer.models.init(self.hub.config['datanommer.sqlalchemy.url'])

    def _load_badges_from_yaml(self, directory):
        # badges indexed by trigger
        badges = []
        directory = os.path.abspath(directory)
        log.info("Looking in %r to load badge definitions" % directory)
        for root, dirs, files in os.walk(directory):
            for partial_fname in files:
                fname = root + "/" + partial_fname
                badge = self._load_badge_from_yaml(fname)

                if not badge:
                    continue

                try:
                    badge_rule = fedbadges.rules.BadgeRule(
                        badge, self.l.tahrir, self.issuer_id)
                    badges.append(badge_rule)
                except ValueError as e:
                    log.error("Initializing rule for %r failed with %r" % (
                        fname, e))

        log.info("Loaded %i total badge definitions" % len(badges))
        return badges

    def _load_badge_from_yaml(self, fname):
        log.debug("Loading %r" % fname)
        try:
            with open(fname, 'r') as f:
                return yaml.load(f.read())
        except Exception as e:
            log.error("Loading %r failed with %r" % (fname, e))
            return None

    def award_badge(self, username, badge_rule, link=None):
        """ A high level way to issue a badge to a Person.

        It adds the person if they don't exist, and creates an assertion for
        them.

        :type username: str
        :param username: This person's username.

        :type badge_rule: object
        :param badge_rule: the badge_rule that triggered this.
        """

        log.info("Awarding badge %r to %r" % (badge_rule.badge_id, username))
        email = "%s@fedoraproject.org" % username

        try:
            transaction.begin()
            self.l.tahrir.add_person(email)
            transaction.commit()
        except:
            transaction.abort()
            self.l.tahrir.session.rollback()
            raise

        user = self.l.tahrir.get_person(email)

        try:
            transaction.begin()
            self.l.tahrir.add_assertion(badge_rule.badge_id, email, None, link)
            transaction.commit()
        except IntegrityError:
            # Here we handle two different kinds of errors due to the existence
            # of a somewhat harmless race condition.
            # Say that someone adds 2 tags to a package in fedora-tagger all at
            # once.  That produces 2 different fedmsg messages that each hit
            # this daemon.  Each message gets handed off to each of 2 worker
            # threads that start working in parallel.  They both ask, does
            # person X have the 'Awesome Tagger' badge?  The database responds
            # "no", so they both start checking to see if the person meets all
            # the criteria.  They do, so the first worker gets here and awards
            # them the badge with add_assertion.  The second thread gets here
            # and tries to award the badge, but it is already awarded by the
            # other thread -- so it raises a 'duplicate primary key'
            # IntegrityError.  We catch that here, and note it in the logs as a
            # warning not an error.  It happens often enough and is harmless
            # enough that we don't want to receive emails about it.
            transaction.abort()
            self.l.tahrir.session.rollback()
            log.warn(traceback.format_exc())
        except:
            # For all other errors, we rollback the transaction but we also
            # re-raise the error so that it hits the fedmsg handling in the
            # stack above and emails us about it.
            transaction.abort()
            self.l.tahrir.session.rollback()
            raise

    def consume(self, msg):

        # First thing, we receive the message, but we put ourselves to sleep to
        # wait for a moment.  The reason for this is that, when things are
        # 'calm' on the bus, we receive messages "too fast".  A message that
        # arrives to the badge awarder triggers (usually) a check against
        # datanommer to count messages.  But if we try to count them before
        # this message arrives at datanommer, we'll get skewed results!  Race
        # condition.  We go to sleep to allow ample time for datanommer to
        # consume this one before we go and start doing checks on it.  When
        # fedbadges was first released, this was absolutely necessary.
        # Since that time, the fedmsg bus has become much more congested.  So,
        # to improve our average speed at handling messages, we only do that
        # sleep statement if we're not already backlogged.  If we know we have
        # a huge workload ahead of us, then go ahead and start handling
        # messages as fast as we can.
        if self.incoming.qsize() < self.delay_limit:
            time.sleep(self.consume_delay)

        # Strip the moksha envelope
        msg = msg['body']

        default = "https://apps.fedoraproject.org/datagrepper"
        link = self.hub.config.get('fedbadges.datagrepper_url', default) + \
            "/id?id=%s&is_raw=true&size=extra-large" % msg['msg_id']

        # Define this so we can refer to it in error handling below
        badge_rule = None

        # Initialize our connection if this is the first time we are called.
        self._initialize_tahrir_connection()

        # Award every badge as appropriate.
        log.debug("Received %s, %s" % (msg['topic'], msg['msg_id']))
        for badge_rule in self.badge_rules:
            try:
                for recipient in badge_rule.matches(msg):
                    self.award_badge(recipient, badge_rule, link)
            except Exception as e:
                log.exception("Rule: %r, message: %r" % (badge_rule, msg))

        log.debug("Done with %s, %s" % (msg['topic'], msg['msg_id']))
